/* Copyright (C) 2010 MoSync AB

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License, version 2, as published by
the Free Software Foundation.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License
along with this program; see the file COPYING.  If not, write to the Free
Software Foundation, 59 Temple Place - Suite 330, Boston, MA
02111-1307, USA.
*/

//This code access a hidden API from Google.  Google may change the specification
//of this API at any time, and it is not guaranteed.

//This is a C++ implementation of the C# code found at http://forum.openhandsetdevelopers.com/google-maps-with-my-location-my-location-t19.html

#include "GoogleCellIDDatabase.h"
#include "../Utilities/Convert.h"
#include <MAUtil/PlaceholderPool.h>
#include <mavsprintf.h>

GoogleCellIDDatabase::GoogleCellIDDatabase()
					: mConnection(this),
					  mData(NULL)
{}

GoogleCellIDDatabase::~GoogleCellIDDatabase()
{
	if(mConnection.isOpen())
		mConnection.close();

	maDestroyObject(mData);
	PlaceholderPool::put(mData);
}

MALocation& GoogleCellIDDatabase::convertToLocation(MACellInfo* cellinfo)
{
	if(!mConnection.isOpen())
	{
		//Create the post data
		getPostData(pd, Convert::hexToInt(cellinfo->mcc), Convert::hexToInt(cellinfo->mnc), cellinfo->lac, cellinfo->cellId, false);

		//Create the connection
		mConnection.create("http://www.google.com/glm/mmap", HTTP_POST);
		mConnection.setRequestHeader("content-length", "55");
		mConnection.setRequestHeader("content-type", "application/binary");

		LOG("Writing the post data");
		//Write the binary request data
		/*String temp;
		for(int i = 0; i < 55; i++)
		{
			char buf[6];
			sprintf(buf, "%d | ", pd[i]);
			temp.append(buf, 5);
		}

		LOG("Raw data follows:");
		lprintfln("%s", temp.c_str());*/
		mConnection.write(&pd, 55);

	}
}

void GoogleCellIDDatabase::connWriteFinished(Connection* conn, int result)
{
	LOG("ConnWriteFinished: %d", result);

	//The post data has been written, so now we need to finish the connection
	LOG("Opening the connection to Google");
	mConnection.finish();
}

void GoogleCellIDDatabase::httpFinished(HttpConnection* http, int result)
{
	//Connection has been established.  Read the result.
	LOG("Http Result: %d", result);
	if(result == 200) //HTTP_OK
	{
		//Get a handle for the data
		if(mData == NULL)
			mData = PlaceholderPool::alloc();

		maCreateData(mData, 512);
		//Reset the receive length counter
		mRecLength = 0;
		mConnection.recvToData(mData, 0, 512); //Read up to 512 bytes of data.  Typical response is 25 bytes
	}
}

void GoogleCellIDDatabase::connRecvFinished(Connection* conn, int result)
{
	//OK, we should have a handle to a data object of our results.
	LOG("Google receive complete");

	//Close the connection
	if(mConnection.isOpen()) mConnection.close();

	//Parse out the data in mData
	//First, copy the data to a byte array
	byte ps[result];
	maReadData(mData, &ps, 0, result);

	//Destroy the data object now
	maDestroyObject(mData);

	LOG("Checking validity");
	//Check the validity of the returned data and parse
	short opcode1 = (short)(ps[0] << 8 | ps[1]);
	byte opcode2 = ps[2];
	if(opcode1 == 0x0e && opcode2 == 0x1b)
	{
	  LOG("OpCode checks complete");
	  int ret_code = (int)((ps[3] << 24) | (ps[4] << 16) | (ps[5] << 8) | (ps[6]));

	  if (ret_code == 0)
	  {
		  LOG("Ret code test passed");
		  double lat = ((double)((ps[7] << 24) | (ps[8] << 16) | (ps[9] << 8) | (ps[10]))) / 1000000;
		  double lon = ((double)((ps[11] << 24) | (ps[12] << 16) | (ps[13] << 8) | (ps[14]))) / 1000000;
		  LOG("Google CellID Result - Latitude: %f, Longitude: %f", lat, lon);
		  mLocation.lat = lat;
		  mLocation.lon = lon;
		  mLocation.horzAcc = 5000; //Assume an accuracy of 5km for a CellID result.

		  //Return the result to location bus
		  APPCONTROLLER->getLocationBus()->receiveLocation(mLocation);
	  }
	  else
		  LOG("Google Error %d", ret_code);
	}

}


void GoogleCellIDDatabase::connectFinished(Connection* conn, int result)
{
	LOG("Connection finished: %d", result);
}

void GoogleCellIDDatabase::connReadFinished(Connection* conn, int result)
{
	LOG("ConnReadFinished: %d", result);
}


void GoogleCellIDDatabase::getPostData(byte* output, int MCC, int MNC, int LAC, int CID, bool shortCID)
{
	/* The shortCID parameter follows heuristic experiences:
	 * Sometimes UMTS CIDs are build up from the original GSM CID (lower 4 hex digits)
	 * and the RNC-ID left shifted into the upper 4 digits.
	 */

	byte pd [] = {
		0x00, 0x0e,
		0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
		0x00, 0x00,
		0x00, 0x00,
		0x00, 0x00,

		0x1b,
		0x00, 0x00, 0x00, 0x00, // Offset 0x11
		0x00, 0x00, 0x00, 0x00, // Offset 0x15
		0x00, 0x00, 0x00, 0x00, // Offset 0x19
		0x00, 0x00,
		0x00, 0x00, 0x00, 0x00, // Offset 0x1f
		0x00, 0x00, 0x00, 0x00, // Offset 0x23
		0x00, 0x00, 0x00, 0x00, // Offset 0x27
		0x00, 0x00, 0x00, 0x00, // Offset 0x2b
		0xff, 0xff, 0xff, 0xff,
		0x00, 0x00, 0x00, 0x00
	};

	bool isUMTSCell = ((long long)CID > 65535);

	if (isUMTSCell)
		LOG("UMTS CID. %s", shortCID ? "Using short CID to resolve." : "");
	else
		LOG("GSM CID given.");

	if (shortCID)
		CID &= 0xFFFF;      /* Attempt to resolve the cell using the GSM CID part */

	if ((long long)CID > 65536) /* GSM: 4 hex digits, UTMS: 6 hex digits */
		pd[0x1c] = 5;
	else
		pd[0x1c] = 3;

	pd[0x11] = (byte)((MNC >> 24) & 0xFF);
	pd[0x12] = (byte)((MNC >> 16) & 0xFF);
	pd[0x13] = (byte)((MNC >> 8) & 0xFF);
	pd[0x14] = (byte)((MNC >> 0) & 0xFF);

	pd[0x15] = (byte)((MCC >> 24) & 0xFF);
	pd[0x16] = (byte)((MCC >> 16) & 0xFF);
	pd[0x17] = (byte)((MCC >> 8) & 0xFF);
	pd[0x18] = (byte)((MCC >> 0) & 0xFF);

	pd[0x27] = (byte)((MNC >> 24) & 0xFF);
	pd[0x28] = (byte)((MNC >> 16) & 0xFF);
	pd[0x29] = (byte)((MNC >> 8) & 0xFF);
	pd[0x2a] = (byte)((MNC >> 0) & 0xFF);

	pd[0x2b] = (byte)((MCC >> 24) & 0xFF);
	pd[0x2c] = (byte)((MCC >> 16) & 0xFF);
	pd[0x2d] = (byte)((MCC >> 8) & 0xFF);
	pd[0x2e] = (byte)((MCC >> 0) & 0xFF);

	pd[0x1f] = (byte)((CID >> 24) & 0xFF);
	pd[0x20] = (byte)((CID >> 16) & 0xFF);
	pd[0x21] = (byte)((CID >> 8) & 0xFF);
	pd[0x22] = (byte)((CID >> 0) & 0xFF);

	pd[0x23] = (byte)((LAC >> 24) & 0xFF);
	pd[0x24] = (byte)((LAC >> 16) & 0xFF);
	pd[0x25] = (byte)((LAC >> 8) & 0xFF);
	pd[0x26] = (byte)((LAC >> 0) & 0xFF);

	//Copy the array to the output array pointer
	memcpy(output, &pd, 55);
}
