/* Copyright (C) 2010 MoSync AB

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License, version 2, as published by
the Free Software Foundation.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License
along with this program; see the file COPYING.  If not, write to the Free
Software Foundation, 59 Temple Place - Suite 330, Boston, MA
02111-1307, USA.
*/

#include "CloudMadePoIProvider.h"
#include "../std.h"
#include "../Utilities/Convert.h"
#include <mavsprintf.h>

CloudMadePoIProvider::CloudMadePoIProvider(const char* apikey)
					: mDownloader(this),
					  mIsBusy(false),
					  mCurrentRequest(NULL),
					  mRequests(2),
					  mApiKey(apikey)
{
	mJSONParser.addListener(this);
}

CloudMadePoIProvider::~CloudMadePoIProvider()
{}

//PoIProvider
void CloudMadePoIProvider::addRequest(PoIRequest* request)
{
	mRequests.enqueue(request);
	process();
}

void CloudMadePoIProvider::addPoIProviderListener(PoIProviderListener* listener)
{
	mListeners.insert(listener);
}

void CloudMadePoIProvider::removePoIProviderListener(PoIProviderListener* listener)
{
	mListeners.erase(listener);
}

//DownloadListener
void CloudMadePoIProvider::finishedDownloading(Downloader* downloader, MAHandle data)
{
	//LOG("CloudMade POI download complete");
	//LOG("Data Size %d bytes", maGetDataSize(data));
	//Read the data into a string
	char tc[maGetDataSize(data) + 1]; //Create a buffer to read the data into
	maReadData(data, &tc, 0, maGetDataSize(data));

	//LOG("Starting JSON parser");
	String temp;
	temp.append(tc, strlen(tc));

	//Reset processing variables
	mObjectCount = 0;
	mFeatureLat = 0;
	mFeatureLon = 0;
	mFeatureName = "";
	mFeatureType = "";
	mGotFeatures = false;

	//Parse the JSON
	mJSONParser.parseString(temp);

	//Clean up the data object
	maDestroyObject(data);

	//Check to see if there is another request in the queue
	process();
}

void CloudMadePoIProvider::notifyProgress(Downloader* downloader, int downloadedBytes, int totalBytes)
{
}

void CloudMadePoIProvider::downloadCancelled(Downloader* downloader)
{}

void CloudMadePoIProvider::error(Downloader* downloader, int code)
{
	LOG("CloudMade POI download error %d - check the url and api key", code);
}

//JSONListener interface
void CloudMadePoIProvider::header(String& header)
{
	//LOG("Header: %s", header.c_str());
	mHeader = header;

	if(mObjectCount == 1 && mHeader == "features")
		mGotFeatures = true;
	else if(mObjectCount == 1)
		mGotFeatures = false;

	if(mGotFeatures && mHeader == "coordinates")
		mCoPos = 1;
	else
		mCoPos = 0;
}

void CloudMadePoIProvider::data(String& value)
{
	//Read latitude and longitude
	switch(mCoPos)
	{
		case 1:
			//The latitude
			//LOG("Setting latitude to %s", value.c_str());
			mPoIResponse.latitude = Convert::toDouble(value);
			mCoPos++;
			break;
		case 2:
			//LOG("Setting longitude to %s", value.c_str());
			mPoIResponse.longitude = Convert::toDouble(value);
			mCoPos = 0;
			break;
	}

	if(mGotFeatures && mHeader == "type")
	{
		//LOG("Setting type to %s", value.c_str());
		mPoIResponse.type = value;
	}
	else if(mGotFeatures && mHeader == "name")
	{
		//LOG("Setting name to %s", value.c_str());
		mPoIResponse.name = value;
	}
}

void CloudMadePoIProvider::arrayStart()
{
}

void CloudMadePoIProvider::arrayEnd()
{}

void CloudMadePoIProvider::objectStart()
{
	mObjectCount++;
	//LOG("mObjectCount %d", mObjectCount);
	if(mObjectCount == 2 && mGotFeatures)
	{
		//LOG("Starting new PoIResponse");
		//This is the start of a new feature, clear old response
		mPoIResponse.latitude = 0;
		mPoIResponse.longitude = 0;
		mPoIResponse.name = "";
		mPoIResponse.type = "";
	}
}

void CloudMadePoIProvider::objectEnd()
{
	mObjectCount--;
	//LOG("mObjectCount %d", mObjectCount);

	if(mObjectCount == 1 && mGotFeatures)
	{
		//LOG("Sending PoIResponse");
		//We've finished with this response, send it to the listeners
		for(Set<PoIProviderListener*>::Iterator itr = mListeners.begin();
			itr != mListeners.end();
			itr++)
		{
			(*itr)->poiFound(mPoIResponse);
		}
	}
}


//Starts to process the next request
void CloudMadePoIProvider::process()
{
	//Check we're not busy and we've got an api key
	if(!mIsBusy && mApiKey.length() > 0)
	{
		if(mRequests.getCount() > 0)
		{
			mIsBusy = true;

			mCurrentRequest = mRequests.dequeue();

			//Format the request url
			char buffer[255];
			//Example URL
			//http://geocoding.cloudmade.com/BC9A493B41014CAABB98F0471D759707/geocoding/v2/find.js?object_type=hotel&around=51.51384,-0.10952&distance=2000&results=10
			sprintf(&buffer[0], "http://geocoding.cloudmade.com/%s/geocoding/v2/find.js?object_type=%s&around=%f,%f&distance=%d&results=%d",
							mApiKey.c_str(),
							mCurrentRequest->type.c_str(),
							mCurrentRequest->latitude,
							mCurrentRequest->longitude,
							mCurrentRequest->radius,
							mCurrentRequest->max);
			LOG("POI - %s", buffer);

			//We can delete this request now
			delete mCurrentRequest;

			//Start the download
			mDownloader.beginDownloading(buffer);
		}
	}
}
