/* Copyright (C) 2010 MoSync AB

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License, version 2, as published by
the Free Software Foundation.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License
along with this program; see the file COPYING.  If not, write to the Free
Software Foundation, 59 Temple Place - Suite 330, Boston, MA
02111-1307, USA.
*/

#include "HttpDownloader.h"
#include "Convert.h"
#include "../std.h"

HttpDownloader::HttpDownloader(DownloadListener* listener)
			  : mConnection(this),
			    mRedirectUrl(NULL),
			    mData(NULL),
			    mTemp(NULL),
			    mSwap(NULL),
			    mConLen(NULL)
{
	mListeners.insert(listener);
}

HttpDownloader::~HttpDownloader()
{
	if(mConnection.isOpen())
		mConnection.close();

	if(mRedirectUrl != NULL)
		delete mRedirectUrl;

	if(mConLen != NULL)
		delete mConLen;

	if(mData != NULL)
	{
		maDestroyObject(mData);
		PlaceholderPool::put(mData);
	}

	if(mTemp != NULL)
	{
		maDestroyObject(mTemp);
		PlaceholderPool::put(mTemp);
	}

	if(mSwap != NULL)
	{
		maDestroyObject(mSwap);
		PlaceholderPool::put(mSwap);
	}
}

int HttpDownloader::beginDownloading(const char* url, MAHandle data)
{
	LOG("Starting HttpDownloader");
	if(mConnection.isOpen())
		mConnection.close();

	LOG("Creating mData");
	//Initialise an MAHandle to receive data to
	if(data > 0)
		mData = data;
	else
		if(mData > 0)
			maDestroyObject(mData);
		else
		{
			LOG("Allocated mData");
			mData = PlaceholderPool::alloc();
		}

	LOG("Reset flags");
	//Reset flags
	mIsRedirect = false;
	mSizeKnown = false;

	//Connect to the url
	LOG("Opening connection");
	int result = mConnection.create(url, HTTP_GET);
	if(result > 0)
		mConnection.finish();

	return result;
}

void HttpDownloader::cancelDownloading()
{
	if(mConnection.isOpen())
		mConnection.close();
}

bool HttpDownloader::isDownloading()
{
	return mConnection.isOpen();
}

//HttpConnectionListener
void HttpDownloader::httpFinished(HttpConnection* http, int result)
{
	LOG("HttpConnection has finished");
	//Http connection has finished, and headers have been sent
	if(result == 301
	|| result == 302
	|| result == 303
	|| result == 307)
	{
		//Process content redirect
		mIsRedirect = true;

		//redirect
		mConnection.close();
		//Get the new url from the response header
		if(mRedirectUrl == NULL)
			mRedirectUrl = new String();
		if(mConnection.getResponseHeader("location", mRedirectUrl) != CONNERR_NOHEADER)
			beginDownloading(mRedirectUrl->c_str(), mData);
		else
			reportError(result); //Report the HTTP error
	}
	else if(result == 200)
	{
		LOG("Found the resource");
		//Process content download
		//Get the content-length header
		if(mConLen == NULL)
			mConLen = new String();
		else
			mConLen->clear();

		if(mConnection.getResponseHeader("content-length", mConLen) != CONNERR_NOHEADER)
		{
			LOG("Got a content-length header");
			mSizeKnown = true;
			//Download all the data
			mSize = Convert::toInt(mConLen->c_str());
			maCreateData(mData, mSize);
			mConnection.readToData(mData, 0, mSize);
		}
		else
		{
			LOG("No content-length header provided");
			bool isOOM = false;
			//Delete any temp data
			if(mTemp != NULL)
				maDestroyObject(mTemp);
			else
			{
				LOG("Creating mTemp");
				mTemp = PlaceholderPool::alloc();
			}

			if(maCreateData(mTemp, 1024) == RES_OUT_OF_MEMORY)
			{
				isOOM = true;
				reportOOM();
				return;
			}

			LOG("Creating mData");
			if(maCreateData(mData, 128) == RES_OUT_OF_MEMORY)
			{
				isOOM = true;
				reportOOM();
				return;
			}
			mPreservemData = false;



			if(!isOOM)
			{
				LOG("Created data");
				mSizeKnown = false;
				//Download 1k segments until no more data exists
				LOG("Starting receive");
				mConnection.recvToData(mTemp, 0, 1024);
			}
		}

	}
	else
	{
		//Report error
		reportError(result);
	}
}

void HttpDownloader::connectFinished(Connection* conn, int result)
{
	//This is called when the initial connection is made, before any request headers
	//have been sent.
}

void HttpDownloader::connRecvFinished(Connection* conn, int result)
{
	LOG("Receive finished.  Got %d bytes", result);
	//This should only be used for unknown sized data
	if(!mSizeKnown)
	{
		LOG("Total size unknown");
		if(result > 0)
		{
			int preservesize = 0; //The number of bytes to preserve

			//We've received some data
			if(mSwap != NULL)
				maDestroyObject(mSwap);
			else
				mSwap = PlaceholderPool::alloc();

			//Prepare to move data
			MACopyData* params = new MACopyData();

			//Copy the contents of mData into mSwap to keep it safe
			if(mPreservemData)
			{
				preservesize = maGetDataSize(mData);
				if(maCreateData(mSwap, maGetDataSize(mData)) == RES_OUT_OF_MEMORY)
				{
					reportOOM();
					return;
				}

				//Move the contents of mData to mSwap
				params->dst = mSwap;
				params->dstOffset = 0;
				params->size = maGetDataSize(mData);
				params->src = mData;
				params->srcOffset = 0;
				maCopyData(params);


				//Resize mData
				maDestroyObject(mData);
				if(maCreateData(mData, maGetDataSize(mSwap) + result) == RES_OUT_OF_MEMORY)
				{
					reportOOM();
					return;
				}

				//Swap data back from mSwap
				params->dst = mData;
				params->src = mSwap;
				maCopyData(params);

			}
			else
			{
				//Nothing to preserve, just resize mData
				maDestroyObject(mData);
				if(maCreateData(mData, result) == RES_OUT_OF_MEMORY)
				{
					reportOOM();
					return;
				}
			}

			//Append the new data from mTemp
			params->src = mTemp;
			params->srcOffset = 0;
			params->dst = mData;
			params->dstOffset = preservesize; //Append at the end of the swapped data
			params->size = result;

			maCopyData(params);

			//Ensure we preserve this data
			mPreservemData = true;

			//All done, clean up
			maDestroyObject(mTemp);
			maDestroyObject(mSwap);
			delete params;

			//Get more data
			if(maCreateData(mTemp, 1024) == RES_OUT_OF_MEMORY)
			{
				reportOOM();
				return;
			}
			else
			{
				reportProgress(result);
				LOG("Getting more data");
				mConnection.recvToData(mTemp, 0, 1024);
			}
		}
		else
		{
			LOG("Reporting finish");
			//Result <= 0.  Assume this is end of data
			reportFinish();
		}
	}
}

void HttpDownloader::connReadFinished(Connection* conn, int result)
{
	//This should only be used when the content length is known.
	reportProgress(result);
	reportFinish();
}

void HttpDownloader::connWriteFinished(Connection* conn, int result)
{

}

void HttpDownloader::reportCancelled()
{
	for(Set<DownloadListener*>::Iterator itr = mListeners.begin();
		itr != mListeners.end();
		itr++)
	{
		(*itr)->downloadCancelled(this);
	}
}

void HttpDownloader::reportError(int errorcode)
{
	for(Set<DownloadListener*>::Iterator itr = mListeners.begin();
		itr != mListeners.end();
		itr++)
	{
		(*itr)->error(this, errorcode);
	}
}

void HttpDownloader::reportFinish()
{
	for(Set<DownloadListener*>::Iterator itr = mListeners.begin();
		itr != mListeners.end();
		itr++)
	{
		(*itr)->finishedDownloading(this, mData);
	}
}

void HttpDownloader::reportOOM()
{
	LOG("Reporting OOM error");
	for(Set<DownloadListener*>::Iterator itr = mListeners.begin();
		itr != mListeners.end();
		itr++)
	{
		(*itr)->outOfMemory(this);
	}
}

void HttpDownloader::reportProgress(int bytes)
{
	for(Set<DownloadListener*>::Iterator itr = mListeners.begin();
		itr != mListeners.end();
		itr++)
	{
		if(mSizeKnown)
			(*itr)->notifyProgress(this, bytes, mSize);
		else
			(*itr)->notifyProgress(this, bytes, -1);
	}
}

