/* Copyright (C) 2010 MoSync AB

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License, version 2, as published by
the Free Software Foundation.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License
along with this program; see the file COPYING.  If not, write to the Free
Software Foundation, 59 Temple Place - Suite 330, Boston, MA
02111-1307, USA.
*/
#include "JSONParser.h"
#include "../std.h"

JSONParser::JSONParser()
{
}

JSONParser::~JSONParser()
{
}

void JSONParser::reset()
{
  _start = 0;
  data.clear();
  temp.clear();
  readingString = false;
}

void JSONParser::parseString(String& jsondata)
{
  data = jsondata;
  _start = 0;
  parse();
}

void JSONParser::parse()
{
  char firstChar;
  while(_start < data.length())
  {
    //Move through white space
    firstChar = data.c_str()[_start];

    while(firstChar < 33)
      firstChar = data.c_str()[++_start];

    //Check the next character, if it is an [,],{ or } raise the appropriate event
    //LOG("First char '%c'", firstChar);
    if(firstChar == '[')
    {
      //LOG("Array Start.  Returning as data '%s'", temp.c_str());
      returnData(temp);

      Vector_each(JSONListener*, itr, _listeners)
        (*itr)->arrayStart();
      _start++;
    }
    else if(firstChar == ']')
    {
      //LOG("Array End.  Returning as data '%s'", temp.c_str());
      returnData(temp);

      Vector_each(JSONListener*, itr, _listeners)
        (*itr)->arrayEnd();
      _start++;
    }
    else if(firstChar == '{')
    {
      //LOG("Object Start.  Returning as data '%s'", temp.c_str());
      returnData(temp);

      Vector_each(JSONListener*, itr, _listeners)
        (*itr)->objectStart();
      _start++;
    }
    else if(firstChar == '}')
    {
      //LOG("Object End.  Returning as data '%s'", temp.c_str());
      returnData(temp);

      Vector_each(JSONListener*, itr, _listeners)
        (*itr)->objectEnd();
      _start++;
    }
    else if(firstChar == ',')
    {
      //LOG("Array Delimiter.  Returning as data '%s'", temp.c_str());
      _start++;
      returnData(temp);
    }
    else if(firstChar == ':')
    {
      //LOG("Header delimiter.  Returning as header '%s'", temp.c_str());
      _start++;
      returnHeader(temp);
    }
    else
    {
      //Process the next section of data, that it until we reach the end of a string, or an object delimiter,
      //or an array delimiter
      temp.clear();

      //If we're reading a string, then the delimiting rules are different
      if(firstChar == '\"')
      {
    	  //Ignore the first quote
    	  _start++;
    	  _end = findNextQuote(_start);
      }
      else
      {
    	  _end = nextDelimiter(_start + 1);
      }

      //LOG("Reading from %d to %d len %d", _start, _start + _end, data.length());
      //Check that a delimiter has been found
      if(_end > _start)
      {
    	  temp = data.substr(_start, _end - _start);
    	  //LOG("I've just read the string '%s'", temp.c_str());
    	  //If we've read a string, then move past the end quote
    	  if(firstChar == '\"') _end++;

    	  //Move to the end of the data we've read
    	  _start = _end;
      }
    }
  }
  //Check that we've not got anything left over
  returnData(temp);
}

void JSONParser::returnHeader(String& header)
{
	//LOG("Returning header: %s", temp.c_str());
	if(header.length() > 0)
		Vector_each(JSONListener*, itr, _listeners)
			(*itr)->header(header);
	header.clear();
}

void JSONParser::returnData(String& data)
{
	if(data.length() > 0)
		Vector_each(JSONListener*, itr, _listeners)
			(*itr)->data(data);
	data.clear();
}

int JSONParser::findNextQuote(int startPos)
{
  startPos = data.findFirstOf('\"', startPos);
  if(data.c_str()[startPos-1] == '\\')
    startPos = findNextQuote(startPos + 1); //Ignore any \" pairs in the string data

	return startPos;
}


void JSONParser::addListener(JSONListener* listener)
{
  _listeners.add(listener);
}

int JSONParser::nextDelimiter(int startPos)
{
  int nextComma = findPosition(',', startPos);
  int nextSquareBrace = findPosition(']', startPos);
  int nextCurlyBrace = findPosition('}', startPos);

  int lowest = nextComma < nextSquareBrace ? nextComma : nextSquareBrace;
  if(nextCurlyBrace < lowest)
    lowest = nextCurlyBrace;

  return lowest;
}

int JSONParser::findPosition(char c, int startPos)
{
  int cpos = data.findFirstOf(c, startPos);
  if(cpos == -1)
    cpos = data.length();

  return cpos;
}

